#promnesia-sidebar.promnesia {
    /* e.g. left should result in: left initial, right 0px, top 0px, bot 0px, height 100%, width 30%
       need to set irrelevant dimensions to 0px, otherwise it misbehaves
    */

    --is-v: var(--left, var(--right ));
    --is-h: var(--top , var(--bottom));

    --init-left  : var(--left  , var(--top , var(--bottom)));
    --init-right : var(--right , var(--top , var(--bottom)));
    --init-top   : var(--top   , var(--left, var(--right )));
    --init-bottom: var(--bottom, var(--left, var(--right )));


    left  : calc(var(--init-left  ) * 0px);
    right : calc(var(--init-right ) * 0px);
    top   : calc(var(--init-top   ) * 0px);
    bottom: calc(var(--init-bottom) * 0px);

    --whelper: calc(var(--is-v) * var(--size));
    --hhelper: calc(var(--is-h) * var(--size));
    width : var(--whelper, 100%);
    height: var(--hhelper, 100%);

    position: fixed;
    z-index: 9999;

    display: none;
}

body.promnesia {
    padding-left  : calc(var(--left)   * var(--size));
    padding-right : calc(var(--right)  * var(--size));
    padding-top   : calc(var(--top)    * var(--size));
    padding-bottom: calc(var(--bottom) * var(--size));
}


/* TODO expose this in settings ? */
.promnesia-highlight {
    background-color: #ffff6688;
}


.nonselectable {
    -webkit-touch-callout: none;
    -webkit-user-select: none;
    -khtml-user-select: none;
    -moz-user-select: none;
    -ms-user-select: none;
    -o-user-select: none;
    user-select: none;
}

.promnesia-highlight-ref {
    color: orange;
    font-weight: bold;
    /* vertical-align: super; */
    margin: 0.1em;
}
