/* @flow */

/*
const doc = document;

import {unwrap, Methods} from './common';
import {defensifyAlert, alertError} from './notifications';

function getInputElement(element_id: string): HTMLInputElement {
    return ((document.getElementById(element_id): any): HTMLInputElement);
}

function getSearch(): HTMLInputElement {
    return getInputElement('search_id');
}


unwrap(getSearch()).addEventListener('click', defensifyAlert(async () => {
    chrome.runtime.sendMessage({ method: Methods.OPEN_SEARCH });
}));

*/
